{*******************************************************************}
{    CpToCp 1.1.1                                                   }
{    Copyright (c) 2000-2002 Andrey Filatkin                        }
{    All Rights Reserved                                            }
{    andrey@filatkin.com ;   http://anfilat.chat.ru/                }
{    Last modified - 17 March 2001                                  }
{*******************************************************************}
{$MINENUMSIZE 4}
library CpToCp;

uses
  Windows, SysUtils, Math, Plugin, FarKeys;

const
  MethodFT     = 0;
  MethodQP     = 1;
  MethodURL    = 2;
  MethodUTF8   = 3;

  MethodFTId   = 1;
  MethodQPId   = 2;
  MethodURLId  = 3;
  MethodUTF8Id = 4;
  FromCpId     = 6;
  ToCpId       = 8;
  Sample1Id    = 10;
  Sample2Id    = 11;
  OkId         = 13;
  CancelId     = 14;

type
  TMessage = (MOk, MCancel, MCpToCp, MCpToCpTitle1, MCpToCpTitle2,
    MFrom, MTo, MWait, MLines, MSample, MWarn, MMesBody);
  TCodeTable = packed array[0..255] of Char;
  TTableName = packed array[0..127] of Char;
  Options = packed record
   Method: Integer;  // 
   FromCp: Integer;  //  
   ToCp: Integer;  // 
  end;
  TTab = record
    cmask, cval, lmask, lval: Integer
  end;
  TTabs = array[0..2] of TTab;

const
  Tabs: TTabs = ((cmask:$80; cval:$00; lmask:$7F; lval:$00),
                 (cmask:$E0; cval:$C0; lmask:$7FF; lval:$80),
                 (cmask:$F0; cval:$E0; lmask:$FFFF; lval:$800));

var
  Info: TPluginStartupInfo;
  PluginMenuStrings: array[0..0] of PChar;
  PluginRootKey: string;
  Opt: Options;
  CountCPs: Integer; //   
  DecodeTables: array of TCodeTable; //   DOS-
  EncodeTables: array of TCodeTable; // DOS-
  TableNames: array of TTableName; //  
  SampleFrom1, SampleFrom2: String; // 
  SampleStart1, SampleEnd1, SampleStart2, SampleEnd2: Integer;//,
    //    
  SampleFromDlg1, SampleFromDlg2: PChar; //  
  EditorCP: Integer;
  ei: TEditorInfo; //

// 
function CreateRegKey(const hRoot: HKey; const Key: string): HKey;
var
  TempKey: HKey;
  Disposition: Integer;
  FullKeyName: string;
begin
  FullKeyName := PluginRootKey;
  if Length(Trim(Key))<>0 then FullKeyName := FullKeyName+'\'+Key;
  RegCreateKeyEx(hRoot,PChar(FullKeyName),0,nil,0,KEY_WRITE,nil,
                 TempKey,@Disposition);
  Result := TempKey;
end;

// 
function OpenRegKey(const hRoot: HKey; const Key: string): HKey;
var
  TempKey: HKey;
  FullKeyName: string;
begin
  FullKeyName := PluginRootKey;
  if Length(Trim(Key))<>0 then FullKeyName := FullKeyName+'\'+Key;
  RegOpenKeyEx(hRoot,PChar(FullKeyName),0,KEY_QUERY_VALUE,TempKey);
  Result := TempKey;
end;

//  
procedure SetRegKey(hRoot: Hkey;Key: string;ValueName: string;ValueData: integer);
var
  TempKey: HKey;
begin
  TempKey := CreateRegKey(hRoot,Key);
  RegSetValueEx(TempKey,PChar(ValueName),0,REG_DWORD,@ValueData,sizeof(ValueData));
  RegCloseKey(TempKey);
end;

//  
function GetRegKey(hRoot: HKey;Key: string;ValueName: string;Default: integer): integer;
var
  ValueData, Type1, Size1, ExitCode : integer;
  TempKey: HKey;
begin
  TempKey := OpenRegKey(hRoot,Key);
  Size1 := sizeof(ValueData);
  ExitCode := RegQueryValueEx(TempKey,PChar(ValueName),nil,@Type1,@ValueData,@Size1);
  RegCloseKey(TempKey);
  if (TempKey=NULL) or (ExitCode<>ERROR_SUCCESS) then ValueData := Default;
  Result := ValueData;
end;

function GetMsg(MsgId: TMessage): PChar;
begin
  result := Info.GetMsg(Info.ModuleNumber, integer(MsgId));
end;

function GetMinFarVersion(): Integer; stdcall;
begin
  result := MakeFarVersion(1,70,1282);
// {1.70b4}    FAR 1.70b4   ()
end;

procedure SetStartupInfo(var psi: TPluginStartupInfo); stdcall;
begin
  Move(psi, Info, SizeOf(Info));
  PluginRootKey := Info.RootKey;
  PluginRootKey := PluginRootKey+'\CpToCp';
end;

procedure GetPluginInfo(var pi: TPluginInfo); stdcall;
begin
  FillChar(pi, SizeOf(pi), 0);
  pi.StructSize := SizeOf(pi);
  pi.Flags := PF_EDITOR or PF_DISABLEPANELS;//    
  PluginMenuStrings[0] := GetMsg(MCpToCp);
  pi.PluginMenuStrings := @PluginMenuStrings;
  pi.PluginMenuStringsNumber := sizeof(PluginMenuStrings) div sizeof(PluginMenuStrings[0]);
end;

// -    -   (  ).
function CheckForKey(vKey: word): boolean;
const
  DELTA = 16;
Type
  TInputRecArr = array[0..Pred(MaxInt div SizeOf(TInputRecord))] of TInputRecord;
Var
  Console: THandle;
  InputRec: ^TInputRecArr;
  InputSize: DWORD;
  ReadSuccess: boolean;
  ReadCnt,i: DWORD;
Begin
  result:= false;
  Console:= GetStdHandle(STD_INPUT_HANDLE);
  InputSize:= 0;
  InputRec:= nil;

  try
    repeat
      Inc(InputSize, DELTA);
      GetMem(InputRec,InputSize*SizeOf(TInputRecord));
      ReadSuccess:= PeekConsoleInput(Console,InputRec^[0],InputSize,ReadCnt);
    until (not ReadSuccess) or (ReadCnt < InputSize);
    if ReadSuccess and (ReadCnt > 0) then
    begin
      for i:= 0 to Pred(ReadCnt) do
        if (InputRec^[i].EventType = KEY_EVENT) and
           (InputRec^[i].Event.KeyEvent.bKeyDown) and
           (InputRec^[i].Event.KeyEvent.wVirtualKeyCode = vKey) then
        begin
         result:= true;
         ReadConsoleInput(Console,InputRec^[0],Succ(i),ReadCnt);
         Break;
        end; { if }
    end; { if }
  finally
    FreeMem(InputRec,0);
  end;
end;

//  
procedure Conversion(var CnvString: PChar; // 
                     const nStart, //   
                     nEnd, // 
                     FromCp, //  
                     ToCp: Integer); // 
var
  i : Integer;
begin
  //    DOS
  if FromCp = 1 then // DOS  WIN?    Windows
   CharToOemBuff(@CnvString[nStart], @CnvString[nStart], nEnd - nStart + 1);
  if FromCp > 1 then // FAR'  
   for i := nStart to nEnd do
     CnvString[i] := Char(DecodeTables[FromCp - 2][Byte(CnvString[i])]);

  //  ,    -
  if ToCp = 1 then
   OemToCharBuff(@CnvString[nStart], @CnvString[nStart], nEnd - nStart + 1);
  if ToCp > 1 then
   for i := nStart to nEnd do
     CnvString[i] := Char(EncodeTables[ToCp - 2][Byte(CnvString[i])]);
end;

//  - 
procedure SampleConversion;
begin
  if Opt.Method = 0 then
  begin
    StrPLCopy(SampleFromDlg1, SampleFrom1, 62); // 62   
      //   
    StrPLCopy(SampleFromDlg2, SampleFrom2, 62);
    if Opt.FromCp <> Opt.ToCp then //  
    begin
      Conversion(SampleFromDlg1, SampleStart1, SampleEnd1, Opt.FromCp, Opt.ToCp);
      Conversion(SampleFromDlg2, SampleStart2, SampleEnd2, Opt.FromCp, Opt.ToCp);
    end;
    if Opt.ToCp <> 0 then //     DOS-  
    begin
      Conversion(SampleFromDlg1, 0, 62, Opt.ToCp, 0);
      Conversion(SampleFromDlg2, 0, 62, Opt.ToCp, 0);
    end
  end
  else
  begin
    StrPLCopy(SampleFromDlg1, '', 0);
    StrPLCopy(SampleFromDlg2, '', 0);
  end;
end;

// 
procedure ChangeSample(hDlg:THandle);
begin
  SampleConversion;  //
  //   
  Info.SendDlgMessage(hDlg, DM_SETTEXTPTR, Sample1Id, Integer(SampleFromDlg1));
  //  
  Info.SendDlgMessage(hDlg, DM_SETTEXTPTR, Sample2Id, Integer(SampleFromDlg2));
end;

//  
function CpToCpDlgProc(hDlg:THandle;Msg:integer;Param1:integer;Param2:integer):integer; stdcall;
var
  i: Integer;
  CurOpt: Options;
  List1: TFarListGetItem;
  res: Integer;
begin
  case Msg of
    DN_EDITCHANGE :
    begin
      CurOpt.FromCp := Opt.FromCp;
      CurOpt.ToCp := Opt.ToCp;
      if Param1 = FromCpId then
        for i := 0 to (CountCPs + 1) do
        begin
          List1.ItemIndex := i;
          Info.SendDlgMessage (hDlg, DM_LISTGETITEM, FromCpId, Integer(@List1));
          if List1.Item.Flags = LIF_SELECTED then Opt.FromCp := i;
        end;
      if Param1 = ToCpId then
        for i := 0 to (CountCPs + 1) do
        begin
          List1.ItemIndex := i;
          Info.SendDlgMessage (hDlg, DM_LISTGETITEM, ToCpId, Integer(@List1));
          if List1.Item.Flags = LIF_SELECTED then Opt.ToCp := i;
        end;
      // 
      if (CurOpt.FromCp <> Opt.FromCp) or (CurOpt.ToCp <> Opt.ToCp) then
        ChangeSample(hDlg); // 
      res:=integer(true);
    end;
    DN_BTNCLICK :
    begin
      res := Integer(true);
      if (Param1 = MethodFTId)and(opt.Method <> MethodFT) then
      begin
        Opt.Method := MethodFT;
        Info.SendDlgMessage(hDlg, DM_ENABLE, FromCpId, Integer(True));
        Info.SendDlgMessage(hDlg, DM_ENABLE, ToCpId, Integer(True));
        ChangeSample(hDlg);
      end;
      if ((Param1 = MethodQPId)or(Param1 = MethodURLId)
       or(Param1 = MethodUTF8Id))and((Opt.Method + 1)<> Param1) then
      begin
        Opt.Method := Param1 - 1;
        Info.SendDlgMessage(hDlg, DM_ENABLE, FromCpId, 0);
        Info.SendDlgMessage(hDlg, DM_ENABLE, ToCpId, 0);
        ChangeSample(hDlg);
      end;
      if Param1 = OkId then res := Integer(false);
      if Param1 = CancelId then res := Integer(false);
    end
    else //    
      res := Info.DefDlgProc(hDlg, Msg, Param1, Param2);
  end;
  result := res;
end;

function QPDecode(var Sres, SQP: string): Boolean;
var
  ch, ch1, ch2 : Char;
  i, j, len, x, y : Integer;
  St : string;
begin
  Result := False;
  len := Length(SQP);
  if len > 0 then
  begin
    ch := SQP[len];
    while (((ch = Char(9)) or (ch = Char(32))) and (len > 0)) do
    begin
      len := len - 1;
      if len > 0 then ch := SQP[len];
    end;
    i := 1;
    j := 1;
    SetLength(St, len);
    while i <= len do
    begin
      ch := SQP[i];
      if ch = '=' then
        if i = len then
        begin
          j := j - 1;
          Result := True;
        end
        else
          if (len - i) > 1 then
          begin
            ch1 := SQP[i + 1];
            ch2 := SQP[i + 2];
            if (((ch1>='0')and(ch1<='9'))or((ch1>='A')and(ch1<='F'))or((ch1>='a')and(ch1<='f')))
             and
             (((ch2>='0')and(ch2<='9'))or((ch2>='A')and(ch2<='F'))or((ch2>='a')and(ch2<='f'))) then
            begin
              if (ch1>='0') and (ch1<='9') then x := Integer(ch1) - Integer('0')
              else x := (Integer(ch1) and $DF) - Integer('A') + 10;
              if (ch2>='0') and (ch2<='9') then y := Integer(ch2) - Integer('0')
              else y := (Integer(ch2) and $DF) - Integer('A') + 10;
              St[j] := Char(x*16 + y);
              i := i + 2;
            end
            else St[j] := ch
          end
          else St[j] := ch
      else St[j] := ch;
      i := i + 1;
      j := j + 1;
    end;
    Sres := Sres + Copy(St, 1, j - 1);
    SetLength(St, 0);
  end;
end;

procedure URLDecode(var CnvString: PChar; const nStart, nEnd, nLen: Integer);
var
  ch, ch1, ch2 : Char;
  i, j, x, y : Integer;
begin
  i := nStart;
  j := nStart;
  while i <= nEnd do
  begin
    ch := CnvString[i];
    if ch = '%' then
      if (nEnd - i) > 1 then
      begin
        ch1 := CnvString[i + 1];
        ch2 := CnvString[i + 2];
        if (((ch1>='0')and(ch1<='9'))or((ch1>='A')and(ch1<='F'))or((ch1>='a')and(ch1<='f')))
         and
         (((ch2>='0')and(ch2<='9'))or((ch2>='A')and(ch2<='F'))or((ch2>='a')and(ch2<='f'))) then
        begin
          if (ch1>='0') and (ch1<='9') then x := Integer(ch1) - Integer('0')
          else x := (Integer(ch1) and $DF) - Integer('A') + 10;
          if (ch2>='0') and (ch2<='9') then y := Integer(ch2) - Integer('0')
          else y := (Integer(ch2) and $DF) - Integer('A') + 10;
          CnvString[j] := Char(x*16 + y);
          i := i + 2;
        end
        else CnvString[j] := ch
      end
      else CnvString[j] := ch
    else CnvString[j] := ch;
    i := i + 1;
    j := j + 1;
  end;
  if nLen > (nEnd + 1) then
    StrMove(CnvString + j, CnvString + i, nLen - nEnd)
  else
    CnvString[j] := char(0);
end;

procedure DeUTF8(var CnvString: PChar; const nStart, nEnd, nLen: Integer);
var
  i, j, k : integer;
  ws : array[0..1] of WideChar;
  ch : array[0..1] of Char;

  function utf8towc(s: pchar): Integer;
  var
    i, l, c0, c, nc : Integer;
  begin
    Result := -1;
    nc := 0;
    c0 := integer(s[0]);
    l := c0;
    for i := 0 to 2 do
    begin
      nc := nc + 1;
      if (c0 and Tabs[i].cmask ) = Tabs[i].cval then
      begin
        l := l and Tabs[i].lmask;
        if ( l < Tabs[i].lval ) then Result := -1
        else begin ws[0] := WideChar(l); Result := nc; end;
        break;
      end;
      s := s + 1;
      c := integer(s[0]);
      if (c and $C0) <> $80 then begin Result := -1; break; end;
      c := c and $7F;
      l := ( l shl 6 ) or c;
    end;
  end;

begin
  i := nStart;
  j := nStart;
  while i <= nEnd do
  begin
    k := utf8towc(CnvString + i);
    if k = -1 then
    begin
      CnvString[j] := CnvString[i];
      i := i + 1;
    end
    else
    begin
      WideCharToMultiByte(CP_ACP,0,@ws,1,@ch,1,nil,nil);
      CnvString[j] := ch[0];
      i := i + k;
    end;
    j := j + 1;
  end;
  if nLen > (nEnd + 1) then
    StrMove(CnvString + j, CnvString + i, nLen - nEnd)
  else
    CnvString[j] := char(0);
end;

procedure DoRecode;
var
  i: Integer;
  hScreen: Integer;
  esp: TEditorSetPosition;
  egs: TEditorGetString;
  ess: TEditorSetString;
  espt: TEditorSetParameter;
  NormalRecode: Boolean;
  MsgItems: array[0..2] of PChar;
  ProgresBar: packed array[0..40] of Char;
  nStep, LinePBStep, CurLine, CountLine, len: Integer;
  NewString: PChar;
  S1, S2, S3 : string;
  ch : Char;
begin
  // -   
  hScreen := Info.SaveScreen(0, 0, -1, -1);

  esp.CurPos        :=  0;
  esp.CurTabPos     := -1;
  esp.TopScreenLine := -1;
  esp.LeftPos       := -1;
  esp.Overtype      := -1;
  egs.StringNumber  := -1;
  ess.StringNumber  := -1;
  // -    
  NormalRecode := True;

  //    
  if ei.BlockType = BTYPE_NONE then
  begin
    CountLine := ei.TotalLines;
    //  -
    MsgItems[0] := GetMsg(MWait);
    //0%  
    FillChar(ProgresBar, 40, 176);
    ProgresBar[40] := Char(0);
    MsgItems[1] := ProgresBar;
    Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
    // - = 40 
    //LinePBStep -     
    // -
    LinePBStep := Trunc(CountLine / 40) - 1;
    if opt.Method = MethodFt then
      for CurLine := 0 to CountLine - 1 do
      begin
        //  Esc -   -  ,
        // 
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        //    (   )
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        //  
        Info.EditorControl(ECTL_GETSTRING, @egs);
        //
        if egs.StringLength <> 0 then
        begin
          NewString := StrNew(egs.StringText);
          Conversion(NewString, 0, (egs.StringLength - 1), Opt.FromCp, Opt.ToCp);

          ess.StringText := NewString;
          ess.StringEOL := egs.StringEOL;
          ess.StringLength := egs.StringLength;
          Info.EditorControl(ECTL_SETSTRING, @ess);

          StrDispose(NewString);
        end;
        if CurLine >= LinePBStep then
        begin
          nStep := Trunc(40 * (CurLine + 1) / CountLine);
          for i := 0 to nStep - 1 do ProgresBar[i] := Char(219);
          LinePBStep := Trunc(((nStep + 1) * CountLine / 40)) - 1;
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
        end
      end
    else if opt.Method = MethodQP then
    begin
      Curline := 0;
      while CurLine < CountLine do
      begin
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        Info.EditorControl(ECTL_GETSTRING, @egs);
        if egs.StringLength <> 0 then
        begin
          S1 := egs.StringText;
          S2 := '';
          while QPDecode(S2, S1) do
          begin
            CountLine := CountLine - 1;
            if CurLine < CountLine then
            begin
              esp.CurLine := CurLine + 1;
              Info.EditorControl(ECTL_SETPOSITION, @esp);
              Info.EditorControl(ECTL_GETSTRING, @egs);
              S1 := egs.StringText;
              Info.EditorControl(ECTL_DELETESTRING, Pointer(0));
              esp.CurLine := CurLine;
              Info.EditorControl(ECTL_SETPOSITION, @esp);
            end
            else
            begin
              S1 := '';
              S2 := S2 + '=';
            end;
          end;

          len := Length(S2);
          if len > 0 then
          begin
            ch := S2[len];
            if (ch = char($0D)) or (ch = char($0A)) then len := len - 1;
          end;

          ess.StringText := PChar(S2);
          ess.StringEOL := egs.StringEOL;
          ess.StringLength := len;
          Info.EditorControl(ECTL_SETSTRING, @ess);
        end;
        if CurLine >= LinePBStep then
        begin
          nStep := Trunc(40 * (CurLine + 1) / CountLine);
          for i := 0 to nStep - 1 do ProgresBar[i] := Char(219);
          LinePBStep := Trunc(((nStep + 1) * CountLine / 40)) - 1;
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
        end;
        CurLine := CurLine + 1;
      end
    end
    else
      for CurLine := 0 to CountLine - 1 do
      begin
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        Info.EditorControl(ECTL_GETSTRING, @egs);
        if egs.StringLength <> 0 then
        begin
          NewString := StrNew(egs.StringText);
          if opt.Method = MethodUTF8 then
            DeUTF8(NewString, 0, (egs.StringLength - 1), egs.StringLength)
          else
            URLDecode(NewString, 0, (egs.StringLength - 1), egs.StringLength);

          ess.StringText := NewString;
          ess.StringEOL := egs.StringEOL;
          ess.StringLength := StrLen(NewString);
          Info.EditorControl(ECTL_SETSTRING, @ess);

          StrDispose(NewString);
        end;
        if CurLine >= LinePBStep then
        begin
          nStep := Trunc(40 * (CurLine + 1) / CountLine);
          for i := 0 to nStep - 1 do ProgresBar[i] := Char(219);
          LinePBStep := Trunc(((nStep + 1) * CountLine / 40)) - 1;
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
        end
      end
  end
  else //    
  begin
    //  -    ,
    //  -   -    
    CurLine := ei.BlockStartLine;
    MsgItems[0] := GetMsg(MWait);
    MsgItems[1] := PChar(GetMsg(MLines) + IntToStr(0));
    Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
    LinePBStep := CurLine + 99;
    if opt.Method = MethodFt then
      while true do
      begin
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        egs.StringNumber := CurLine;
        if Info.EditorControl(ECTL_GETSTRING, @egs) = 0 then break;
        if egs.SelStart = -1 then break;
        if (egs.SelEnd = -1) or (egs.SelEnd > egs.StringLength) then
        begin
          egs.SelEnd := egs.StringLength;
          if egs.SelEnd < egs.SelStart then egs.SelEnd := egs.SelStart;
        end;
        NewString := StrNew(egs.StringText);
        Conversion(NewString, egs.SelStart, egs.SelEnd - 1, Opt.FromCp, Opt.ToCp);

        ess.StringText := NewString;
        ess.StringEOL := egs.StringEOL;
        ess.StringLength := egs.StringLength;
        Info.EditorControl(ECTL_SETSTRING, @ess);

        StrDispose(NewString);
        CurLine := CurLine + 1;
        //    
        if CurLine >= LinePBStep then
        begin
          MsgItems[1] := PChar(GetMsg(MLines) +
             IntToStr(LinePBStep - ei.BlockStartLine + 1));
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
          LinePBStep := LinePBStep + 100;
        end
      end
    else if opt.Method = MethodQP then
      while true do
      begin
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        egs.StringNumber := CurLine;
        if Info.EditorControl(ECTL_GETSTRING, @egs) = 0 then break;
        if egs.SelStart = -1 then break;
        if (egs.SelEnd = -1) or (egs.SelEnd > egs.StringLength) then
        begin
          egs.SelEnd := egs.StringLength;
          if egs.SelEnd < egs.SelStart then egs.SelEnd := egs.SelStart;
        end;
        if egs.SelEnd <> egs.SelStart then
        begin
          S1 := Copy(egs.StringText, egs.SelStart + 1, (egs.SelEnd - egs.SelStart));
          S2 := Copy(egs.StringText, 1, egs.SelStart);
          S3 := Copy(egs.StringText, egs.SelEnd + 1, (egs.StringLength - egs.SelEnd));
          while QPDecode(S2, S1) do
            if S3 <> '' then
            begin S1 := ''; S2 := S2 + '='; end
            else
            begin
              esp.CurLine := CurLine + 1;
              Info.EditorControl(ECTL_SETPOSITION, @esp);
              egs.StringNumber := CurLine + 1;
              if Info.EditorControl(ECTL_GETSTRING, @egs) = 0 then
              begin S1 := ''; S2 := S2 + '='; end
              else
                if (egs.SelStart = 0) and (egs.SelEnd <> 0) then
                begin
                  if egs.SelEnd = -1 then S1 := egs.StringText
                  else
                  begin
                    S1 := Copy(egs.StringText, 1, egs.SelEnd);
                    S3 := Copy(egs.StringText, egs.SelEnd + 1, (egs.StringLength - egs.SelEnd));
                  end;
                  Info.EditorControl(ECTL_DELETESTRING, Pointer(0));
                end
                else begin S1 := ''; S2 := S2 + '='; end;
              esp.CurLine := CurLine;
              Info.EditorControl(ECTL_SETPOSITION, @esp);
            end;

          S2 := S2 + S3;
          len := Length(S2);
          if len > 0 then
          begin
            ch := S2[len];
            if (ch = char($0D)) or (ch = char($0A)) then len := len - 1;
          end;

          ess.StringText := PChar(S2);
          ess.StringEOL := egs.StringEOL;
          ess.StringLength := len;
          Info.EditorControl(ECTL_SETSTRING, @ess);
        end;
        CurLine := CurLine + 1;
        if CurLine >= LinePBStep then
        begin
          MsgItems[1] := PChar(GetMsg(MLines) +
             IntToStr(LinePBStep - ei.BlockStartLine + 1));
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
          LinePBStep := LinePBStep + 100;
        end
      end
    else
      while true do
      begin
        if CheckForKey(VK_ESCAPE) then
        begin
          NormalRecode := false;
          break;
        end;
        esp.CurLine := CurLine;
        Info.EditorControl(ECTL_SETPOSITION, @esp);
        egs.StringNumber := CurLine;
        if Info.EditorControl(ECTL_GETSTRING, @egs) = 0 then break;
        if egs.SelStart = -1 then break;
        if (egs.SelEnd = -1) or (egs.SelEnd > egs.StringLength) then
        begin
          egs.SelEnd := egs.StringLength;
          if egs.SelEnd < egs.SelStart then egs.SelEnd := egs.SelStart;
        end;
        NewString := StrNew(egs.StringText);
        if opt.Method = MethodUTF8 then
          DeUTF8(NewString, egs.SelStart, egs.SelEnd - 1, egs.StringLength)
        else
          URLDecode(NewString, egs.SelStart, egs.SelEnd - 1, egs.StringLength);

        ess.StringText := NewString;
        ess.StringEOL := egs.StringEOL;
        ess.StringLength := StrLen(NewString);
        Info.EditorControl(ECTL_SETSTRING, @ess);

        StrDispose(NewString);
        CurLine := CurLine + 1;
        if CurLine >= LinePBStep then
        begin
          MsgItems[1] := PChar(GetMsg(MLines) +
             IntToStr(LinePBStep - ei.BlockStartLine + 1));
          Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 2, 0);
          LinePBStep := LinePBStep + 100;
        end
      end
  end;
  //    
  esp.CurLine       := ei.CurLine;
  esp.CurPos        := ei.CurPos;
  esp.TopScreenLine := ei.TopScreenLine;
  esp.LeftPos       := ei.LeftPos;
  Info.EditorControl(ECTL_SETPOSITION, @esp);

  Info.RestoreScreen(hScreen);

  //    
  if NormalRecode then
  begin
    espt.ParamType := ESPT_CHARTABLE;
    if (Opt.Method = MethodFT) and (EditorCP <> Opt.ToCp) then
    begin
      espt.Param.iParam := Opt.ToCp + 1;
      Info.EditorControl(ECTL_SETPARAM, @espt);
    end;
    if (Opt.Method = MethodUTF8) and (EditorCP <> 1) then
    begin
      espt.Param.iParam := 2;
      Info.EditorControl(ECTL_SETPARAM, @espt);
    end;
  end
  else //    -
  begin //   
    MsgItems[0] := GetMsg(MWarn);
    MsgItems[1] := GetMsg(MMesBody);
    MsgItems[2] := GetMsg(MOk);
    Info.Message(Info.ModuleNumber, 0, NIL, @MsgItems, 3, 1);
  end;
end;

//   
function OpenPlugin(OpenFrom: integer; Item: integer): THandle; stdcall;
var
  cts1: TCharTableSet;
  i, j, ExitCode: Integer;
  DialogItems: array[0..14] of TFarDialogItem;
  FarList1, FarList2: TFarList;
  FarListItemArr1, FarListItemArr2: Pointer;
  TempPointer: Pointer;
  FarListItem1: TFarListItem;
  egs: TEditorGetString;
begin
  //   
  Info.EditorControl(ECTL_GETINFO, @ei);
  //OEM -  0-  
  //ANSI - 1-
  //FAR'   - 2-  
  if (ei.TableNum = -1) then
   if ei.AnsiMode then Opt.FromCp := 1 else Opt.FromCp := 0
  else Opt.FromCp := ei.TableNum + 2;
  //  
  EditorCP := Opt.FromCp;
  Opt.Method := GetRegKey(HKEY_CURRENT_USER,'','Method',MethodFT);
  Opt.ToCp := GetRegKey(HKEY_CURRENT_USER,'','ToCp',0);//
    //      

  i := 0;
  //   
  while (Info.CharTable(i, @cts1, sizeof(cts1)) <> -1) do i := i + 1;
  CountCPs := i;
  //         OEM
  SetLength(DecodeTables, CountCPs);
  // OEM
  SetLength(EncodeTables, CountCPs);
  //  
  SetLength(TableNames, CountCPs);
  //  
  for j := 0 to (CountCPs - 1) do
  begin
    Info.CharTable(j, @cts1, sizeof(cts1));
    Move(cts1.DecodeTable, DecodeTables[j], 256);
    Move(cts1.EncodeTable, EncodeTables[j], 256);
    Move(cts1.TableName, TableNames[j], 128);
  end;
  //    ?
  if (Opt.ToCp > (CountCPs + 1)) then Opt.ToCp := (CountCPs + 1);

  //     ( )

  // -       ?
  //       
  //  
  GetMem(FarListItemArr1, sizeof(TFarListItem) * (CountCPs + 2));
  GetMem(FarListItemArr2, sizeof(TFarListItem) * (CountCPs + 2));
  //   (FarListItem1)  
  //  
  if Opt.FromCp = 0 then FarListItem1.Flags := LIF_SELECTED
  else FarListItem1.Flags := 0;
  StrCopy(FarListItem1.Text, 'DOS (OEM)');
  // -    
  TempPointer := Pointer(Integer(FarListItemArr1));
  //     (      )
  Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
  // 
  if Opt.ToCp = 0 then FarListItem1.Flags := LIF_SELECTED
  else FarListItem1.Flags := 0;
  TempPointer := Pointer(Integer(FarListItemArr2));
  Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
  //   
  if Opt.FromCp = 1 then FarListItem1.Flags := LIF_SELECTED
  else FarListItem1.Flags := 0;
  StrCopy(FarListItem1.Text, 'WIN (ANSI)');
  TempPointer := Pointer(Integer(FarListItemArr1)+sizeof(TFarListItem));
  Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
  if Opt.ToCp = 1 then FarListItem1.Flags := LIF_SELECTED
  else FarListItem1.Flags := 0;
  TempPointer := Pointer(Integer(FarListItemArr2)+sizeof(TFarListItem));
  Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
  // 
  for j := 0 to (CountCPs - 1) do
  begin
    if Opt.FromCp = (j + 2) then FarListItem1.Flags := LIF_SELECTED
    else FarListItem1.Flags := 0;
    StrCopy(FarListItem1.Text, TableNames[j]);
    TempPointer := Pointer(Integer(FarListItemArr1)+sizeof(TFarListItem)*(j+2));
    Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
    if Opt.ToCp = (j + 2) then FarListItem1.Flags := LIF_SELECTED
    else FarListItem1.Flags := 0;
    TempPointer := Pointer(Integer(FarListItemArr2)+sizeof(TFarListItem)*(j+2));
    Move(FarListItem1, TempPointer^, sizeof(TFarListItem));
  end;

  //  
  //  
  if ei.BlockType = BTYPE_NONE then
  begin
    // 
    egs.StringNumber := 0;
    Info.EditorControl(ECTL_GETSTRING, @egs);
    SampleFrom1 := StrNew(egs.StringText);
    SampleStart1 := 0;
    SampleEnd1 := Min((egs.StringLength - 1), 61);
    // 
    if ei.TotalLines > 1 then
    begin
      egs.StringNumber := 1;
      Info.EditorControl(ECTL_GETSTRING, @egs);
      SampleFrom2 := StrNew(egs.StringText);
      SampleStart2 := 0;
      SampleEnd2 := Min((egs.StringLength - 1), 61);
    end
    else
    begin
      SampleFrom2 := PChar('');
      SampleStart2 := 0;
      SampleEnd2 := 0;
    end;
  end
  else // 
  begin
    egs.StringNumber := ei.BlockStartLine;
    Info.EditorControl(ECTL_GETSTRING, @egs);
    SampleFrom1 := StrNew(egs.StringText);
    if (egs.SelEnd = -1) or (egs.SelEnd > egs.StringLength) then
    begin
      egs.SelEnd := egs.StringLength;
      if egs.SelEnd < egs.SelStart then egs.SelEnd := egs.SelStart;
    end;
    SampleStart1 := egs.SelStart;
    SampleEnd1 := Min((egs.SelEnd - 1), 61);
    if SampleEnd1 < SampleStart1 then SampleEnd1 := SampleStart1;
    egs.StringNumber := ei.BlockStartLine + 1;
    if (Info.EditorControl(ECTL_GETSTRING, @egs) = 0)
      or (egs.SelStart = -1) or (egs.SelStart = egs.SelEnd) then
    begin
      SampleFrom2 := PChar('');
      SampleStart2 := 0;
      SampleEnd2 := 0;
    end
    else
    begin
      SampleFrom2 := StrNew(egs.StringText);
      if (egs.SelEnd = -1) or (egs.SelEnd > egs.StringLength) then
      begin
        egs.SelEnd := egs.StringLength;
        if egs.SelEnd < egs.SelStart then egs.SelEnd := egs.SelStart;
      end;
      SampleStart2 := egs.SelStart;
      SampleEnd2 := Min((egs.SelEnd - 1), 61);
      if SampleEnd2 < SampleStart2 then SampleEnd2 := SampleStart2;
    end;
  end;
  // 
  SampleFromDlg1 := StrAlloc(63);
  SampleFromDlg2 := StrAlloc(63);
  SampleConversion;

  //  
  DialogItems[0].ItemType := DI_DOUBLEBOX;
  DialogItems[0].X1 := 3;
  DialogItems[0].Y1 := 1;
  DialogItems[0].X2 := 70;
  DialogItems[0].Y2 := 13;
  DialogItems[0].Focus := 0;
  DialogItems[0].Param.Selected := False;
  DialogItems[0].Flags := DIF_BOXCOLOR;
  DialogItems[0].DefaultButton := False;
  if ei.BlockType=BTYPE_NONE then
    StrCopy(DialogItems[0].Data.Data, GetMsg(MCpToCpTitle2))
  else StrCopy(DialogItems[0].Data.Data, GetMsg(MCpToCpTitle1));

  DialogItems[MethodFTId].ItemType := DI_RADIOBUTTON;
  DialogItems[MethodFTId].X1 := 4;
  DialogItems[MethodFTId].Y1 := 2;
  DialogItems[MethodFTId].X2 := 0;
  DialogItems[MethodFTId].Y2 := 0;
  DialogItems[MethodFTId].Focus := 0;
  DialogItems[MethodFTId].Param.Selected := False;
  DialogItems[MethodFTId].Flags := DIF_GROUP;
  DialogItems[MethodFTId].DefaultButton := False;
  StrCopy(DialogItems[MethodFTId].Data.Data, 'Far Tables');

  DialogItems[MethodQPId].ItemType := DI_RADIOBUTTON;
  DialogItems[MethodQPId].X1 := 4;
  DialogItems[MethodQPId].Y1 := 9;
  DialogItems[MethodQPId].X2 := 0;
  DialogItems[MethodQPId].Y2 := 0;
  DialogItems[MethodQPId].Focus := 0;
  DialogItems[MethodQPId].Param.Selected := False;
  DialogItems[MethodQPId].Flags := 0;
  DialogItems[MethodQPId].DefaultButton := False;
  StrCopy(DialogItems[MethodQPId].Data.Data, 'Quoted-Printable');

  DialogItems[MethodURLId].ItemType := DI_RADIOBUTTON;
  DialogItems[MethodURLId].X1 := 37;
  DialogItems[MethodURLId].Y1 := 9;
  DialogItems[MethodURLId].X2 := 0;
  DialogItems[MethodURLId].Y2 := 0;
  DialogItems[MethodURLId].Focus := 0;
  DialogItems[MethodURLId].Param.Selected := False;
  DialogItems[MethodURLId].Flags := 0;
  DialogItems[MethodURLId].DefaultButton := False;
  StrCopy(DialogItems[MethodURLId].Data.Data, 'URL format');

  DialogItems[MethodUTF8Id].ItemType := DI_RADIOBUTTON;
  DialogItems[MethodUTF8Id].X1 := 4;
  DialogItems[MethodUTF8Id].Y1 := 10;
  DialogItems[MethodUTF8Id].X2 := 0;
  DialogItems[MethodUTF8Id].Y2 := 0;
  DialogItems[MethodUTF8Id].Focus := 0;
  DialogItems[MethodUTF8Id].Param.Selected := False;
  DialogItems[MethodUTF8Id].Flags := 0;
  DialogItems[MethodUTF8Id].DefaultButton := False;
  StrCopy(DialogItems[MethodUTF8Id].Data.Data, 'UTF-8');

  DialogItems[Opt.Method + 1].Focus := 1;
  DialogItems[Opt.Method + 1].Param.Selected := True;

  DialogItems[5].ItemType := DI_TEXT;
  DialogItems[5].X1 := 6;
  DialogItems[5].Y1 := 3;
  DialogItems[5].X2 := 0;
  DialogItems[5].Y2 := 0;
  DialogItems[5].Focus := 0;
  DialogItems[5].Param.Selected := False;
  DialogItems[5].Flags := 0;
  DialogItems[5].DefaultButton := False;
  StrCopy(DialogItems[5].Data.Data, GetMsg(MFrom));

  FarList1.ItemsNumber := CountCPs + 2;
  FarList1.Items := FarListItemArr1;

  DialogItems[FromCpId].ItemType := DI_COMBOBOX;
  DialogItems[FromCpId].X1 := 6;
  DialogItems[FromCpId].Y1 := 4;
  DialogItems[FromCpId].X2 := 35;
  DialogItems[FromCpId].Y2 := 0;
  DialogItems[FromCpId].Focus := 0;
  DialogItems[FromCpId].Param.ListItems := @FarList1;
  DialogItems[FromCpId].Flags := DIF_DROPDOWNLIST or DIF_LISTWRAPMODE
   or DIF_LISTAUTOHIGHLIGHT;
  if Opt.Method <> MethodFT then
   DialogItems[FromCpId].Flags := DialogItems[FromCpId].Flags or DIF_DISABLE;
  DialogItems[FromCpId].DefaultButton := False;
  //  Data   
  StrCopy(DialogItems[FromCpId].Data.Data, Chr(0));

  DialogItems[7].ItemType := DI_TEXT;
  DialogItems[7].X1 := 39;
  DialogItems[7].Y1 := 3;
  DialogItems[7].X2 := 0;
  DialogItems[7].Y2 := 0;
  DialogItems[7].Focus := 0;
  DialogItems[7].Param.Selected := False;
  DialogItems[7].Flags := 0;
  DialogItems[7].DefaultButton := False;
  StrCopy(DialogItems[7].Data.Data, GetMsg(MTo));

  FarList2.ItemsNumber := CountCPs + 2;
  FarList2.Items := FarListItemArr2;

  DialogItems[ToCpId].ItemType := DI_COMBOBOX;
  DialogItems[ToCpId].X1 := 39;
  DialogItems[ToCpId].Y1 := 4;
  DialogItems[ToCpId].X2 := 68;
  DialogItems[ToCpId].Y2 := 0;
  DialogItems[ToCpId].Focus := 0;
  DialogItems[ToCpId].Param.ListItems := @FarList2;
  DialogItems[ToCpId].Flags := DIF_DROPDOWNLIST or DIF_LISTWRAPMODE
   or DIF_LISTAUTOHIGHLIGHT;
  if Opt.Method <> MethodFT then
   DialogItems[ToCpId].Flags := DialogItems[ToCpId].Flags or DIF_DISABLE;
  DialogItems[ToCpId].DefaultButton := False;
  StrCopy(DialogItems[ToCpId].Data.Data, Chr(0));

  DialogItems[9].ItemType := DI_SINGLEBOX;
  DialogItems[9].X1 := 6;
  DialogItems[9].Y1 := 5;
  DialogItems[9].X2 := 69;
  DialogItems[9].Y2 := 8;
  DialogItems[9].Focus := 0;
  DialogItems[9].Param.Selected := False;
  DialogItems[9].Flags := DIF_BOXCOLOR;
  DialogItems[9].DefaultButton := False;
  StrCopy(DialogItems[9].Data.Data, GetMsg(MSample));

  //  
  DialogItems[Sample1Id].ItemType := DI_TEXT;
  DialogItems[Sample1Id].X1 := 7;
  DialogItems[Sample1Id].Y1 := 6;
  DialogItems[Sample1Id].X2 := 0;
  DialogItems[Sample1Id].Y2 := 0;
  DialogItems[Sample1Id].Focus := 0;
  DialogItems[Sample1Id].Param.Selected := False;
  DialogItems[Sample1Id].Flags := DIF_SHOWAMPERSAND;
  DialogItems[Sample1Id].DefaultButton := False;
  StrLCopy(DialogItems[Sample1Id].Data.Data, SampleFromDlg1, 62);

  //  
  DialogItems[Sample2Id].ItemType := DI_TEXT;
  DialogItems[Sample2Id].X1 := 7;
  DialogItems[Sample2Id].Y1 := 7;
  DialogItems[Sample2Id].X2 := 0;
  DialogItems[Sample2Id].Y2 := 0;
  DialogItems[Sample2Id].Focus := 0;
  DialogItems[Sample2Id].Param.Selected := False;
  DialogItems[Sample2Id].Flags := DIF_SHOWAMPERSAND;
  DialogItems[Sample2Id].DefaultButton := False;
  StrLCopy(DialogItems[Sample2Id].Data.Data, SampleFromDlg2, 62);


  DialogItems[12].ItemType := DI_TEXT;
  DialogItems[12].X1 := 0;
  DialogItems[12].Y1 := 11;
  DialogItems[12].X2 := 0;
  DialogItems[12].Y2 := 0;
  DialogItems[12].Focus := 0;
  DialogItems[12].Param.Selected := False;
  DialogItems[12].Flags := DIF_BOXCOLOR or DIF_SEPARATOR;
  DialogItems[12].DefaultButton := False;
  StrCopy(DialogItems[12].Data.Data, '');

  DialogItems[OkId].ItemType := DI_BUTTON;
  DialogItems[OkId].X1 := 0;
  DialogItems[OkId].Y1 := 12;
  DialogItems[OkId].X2 := 0;
  DialogItems[OkId].Y2 := 0;
  DialogItems[OkId].Focus := 0;
  DialogItems[OkId].Param.Selected := False;
  DialogItems[OkId].Flags := DIF_CENTERGROUP;
  DialogItems[OkId].DefaultButton := True;
  StrCopy(DialogItems[OkId].Data.Data, GetMsg(MOk));

  DialogItems[CancelId].ItemType := DI_BUTTON;
  DialogItems[CancelId].X1 := 0;
  DialogItems[CancelId].Y1 := 12;
  DialogItems[CancelId].X2 := 0;
  DialogItems[CancelId].Y2 := 0;
  DialogItems[CancelId].Focus := 0;
  DialogItems[CancelId].Param.Selected := False;
  DialogItems[CancelId].Flags := DIF_CENTERGROUP;
  DialogItems[CancelId].DefaultButton := False;
  StrCopy(DialogItems[CancelId].Data.Data, GetMsg(MCancel));

  //
  ExitCode := Info.DialogEx(Info.ModuleNumber, -1, -1, 74, 15, 'Contents',
                        @DialogItems, 15, 0, 0, CpToCpDlgProc, 0);

  //        :-)
  FreeMem(FarListItemArr1);
  FreeMem(FarListItemArr2);
  StrDispose(SampleFromDlg1);
  StrDispose(SampleFromDlg2);

  //  
  if ExitCode = OkId then
  begin
    SetRegKey(HKEY_CURRENT_USER, '', 'ToCp', Opt.ToCp);
    SetRegKey(HKEY_CURRENT_USER, '', 'Method', Opt.Method);

    if ((opt.Method = MethodFT) and (Opt.FromCp <> Opt.ToCp))
     or (opt.Method = MethodQP) or (opt.Method = MethodUTF8)
     or (opt.Method = MethodURL) then DoRecode;
  end;
  Result := INVALID_HANDLE_VALUE;
end;

exports
  SetStartupInfo,
  GetPluginInfo,
  OpenPlugin,
  GetMinFarVersion;
end.
